/*
   Copyright (c) 2004-2005 The University of Chicago, as Operator
   of Argonne National Laboratory.
*/
package gov.anl.aps.irmis.apps.componenttype.cfw;

import java.net.URL;
import java.util.List;
import java.util.Set;
import java.util.Iterator;
import javax.swing.SwingUtilities;
import javax.swing.JDesktopPane;
import javax.swing.JFrame;

// XAL framework stuff
import gov.sns.application.*;

// general app utilities
import gov.anl.aps.irmis.apps.SwingThreadWork;
import gov.anl.aps.irmis.apps.QueuedExecutorUtil;

// persistence layer
import gov.anl.aps.irmis.persistence.component.ComponentTypeDAO;
import gov.anl.aps.irmis.persistence.component.ComponentType;
import gov.anl.aps.irmis.persistence.DAOException;

// service layer
import gov.anl.aps.irmis.service.IRMISException;
import gov.anl.aps.irmis.service.component.ComponentService;


/**
 * The primary controller class for IRMIS Component Type Viewer/Editor application.
 * Provides "action" methods to to query the IRMIS database for component type
 * related data. These action methods are typically invoked by the 
 * <code>ComponentTypeWindow</code>.
 * Any time-consuming activity should be queued up using the 
 * <code>QueuedExecutor</code>, which is similar to the SwingWorker concept.
 */
public class ComponentTypeDocument extends XalInternalDocument {

	/** The main data model of this application */
	private ComponentTypeModel _model;

    // Hold reference to parent desktop frame. Used to put up busy icon during work.
    JFrame _appFrame;

	/** Create a new empty document */
    public ComponentTypeDocument() {
        this(null);
    }

	/** Create a new document from url */
    public ComponentTypeDocument(final URL url) {
        //setSource(url);
        _model = new ComponentTypeModel();
        _appFrame = ((DesktopApplication)Application.getApp()).getDesktopFrame();
        // don't have anything here yet
    }

    public void saveDocumentAs(final URL url) {
        return;
    }

    /**
     * Specifies the document file suffixes supported by this document.
     * @return An array of file suffixes corresponding to writable files
     */
    public String[] writableDocumentTypes() {
        return new String[] {};
	}

    /**
	 * Overriding default window title to my liking.
     * @param newTitle The new title for this document.
     */
    public void setTitle( final String newTitle ) {
        String modifiedTitle = "idtcomponenttype" + newTitle;
		super.setTitle( modifiedTitle );
        _documentListenerProxy.titleChanged( this, title );
    }	

    /**
     * Queue up job for background thread to gather the current set of ComponentType
     * objects. When complete, fire off <code>ComponentTypeModelEvent</code> to notify GUI
     * that new list of comp types is available for display. 
     */
    public void loadInitialComponentTypeData() {

        QueuedExecutorUtil.execute(new SwingThreadWork(_appFrame) {
                private List compTypeList = null;
                //private List iocAttrTypeList = null;

                public void doNonUILogic() {

                    ComponentService componentService = new ComponentService();
                    // Get current set of ComponentType objects
                    try {
                        // get list of comp types
                        compTypeList = componentService.findComponentTypeList();

                        // get list of possible attribute types
                        //iocAttrTypeList = pvService.findIOCAttributeTypeList();
                        
                    } catch (Exception e) {
                        e.printStackTrace();
                        // log error
                    }
                }
                
                public void doUIUpdateLogic() {
                    // fire off notification of modification to listeners
                    _model.setComponentTypeList(compTypeList);
                    _model.notifyComponentTypeModelListeners(new ComponentTypeModelEvent(ComponentTypeModelEvent.NEW_COMPONENT_TYPE_LIST));
                }
            });
    }
            
    /**
     * Make a main window by instantiating the my custom window.
     */
    public void makeMainWindow() {
        _mainWindow = new ComponentTypeWindow(this);
        loadInitialComponentTypeData();
    }


	/**
	 * Get the launch model which represents the main model of this document
	 * @return the main model of this document
	 */
	public ComponentTypeModel getModel() {
		return _model;
	}

    private ComponentTypeWindow myWindow() {
        return (ComponentTypeWindow)_mainWindow;
    }
}
