/*
   Copyright (c) 2004-2005 The University of Chicago, as Operator
   of Argonne National Laboratory.
*/
package gov.anl.aps.irmis.apps.componenttype.cfw;

import java.io.File;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.List;
import java.util.Iterator;
import java.util.Enumeration;

import java.awt.*;
import java.awt.event.*;

import javax.swing.*;
import javax.swing.table.*;
import javax.swing.border.TitledBorder;
import javax.swing.event.ListSelectionListener;
import javax.swing.event.ListSelectionEvent;

import gov.sns.application.*;

// persistence layer
import gov.anl.aps.irmis.persistence.component.ComponentType;

// for registering protected gui components
import gov.anl.aps.irmis.login.LoginUtil;
import gov.anl.aps.irmis.login.RolePrincipal;

/**
 * Primary GUI for IRMIS Component Type Viewer/Editor application. 
 * Business logic work is delegated to <code>ComponentTypeDocument</code>, which
 * in turn requests that the ComponentTypeModel notify us here of changes to the data. In short,
 * we listen for <code>ComponentTypeModelEvent</code> here.
 */
public class ComponentTypeWindow extends XalInternalWindow {

	/** The main model for the document */
	final protected ComponentTypeModel _model;
    final protected ComponentTypeDocument document;

    static RolePrincipal[] adminPrincipal = { new RolePrincipal("admin") };

	// Swing GUI components
    private JPanel topPanel;
    private JPanel compTypesPanel;
    private JPanel compTypeListPanel;
    private JPanel selectedCompTypePanel;
    private JPanel compTypeDetailsPanel;

    private JSplitPane lrSplitPane;

    private JTable compTypeList;

    private JButton compTypeEditButton;
    private JButton compTypeAddButton;
	
	/** 
	 * Creates a new instance of ComponentWindow
	 * @param aDocument The document for this window
	 */
    public ComponentTypeWindow(final XalInternalDocument aDocument) {
        super(aDocument);
        document = (ComponentTypeDocument)aDocument;

		_model = ((ComponentTypeDocument)aDocument).getModel();

        // make ComponentTypeWindow a listener for changes in ComponentTypeModel
		_model.addComponentTypeModelListener( new ComponentTypeModelListener() {
                public void modified(ComponentTypeModelEvent e) {
                    updateView(e);
                }
            });

        // initial application window size
        setSize(420, 205);

        // build contents
		makeContents();
    }

    /**
     * Top-level method to build up Swing GUI components.
     */
    public void makeContents() {

        // scrollbar policy
        int hsbp = JScrollPane.HORIZONTAL_SCROLLBAR_NEVER;
        int vsbp = JScrollPane.VERTICAL_SCROLLBAR_ALWAYS;

        // topmost panel
        topPanel = new JPanel();
        topPanel.setLayout(new GridBagLayout());
        GridBagConstraints c = new GridBagConstraints();
        c.weightx = 1.0;
        c.weighty = 1.0;
        c.fill = GridBagConstraints.BOTH;
        c.gridx = 0;
        c.gridy = 0;

        compTypesPanel = new JPanel();
        compTypesPanel.setLayout(new BoxLayout(compTypesPanel,BoxLayout.PAGE_AXIS));

        // comp type list panel
        compTypeListPanel = new JPanel(new BorderLayout());
        TitledBorder compTypeListTitle = BorderFactory.createTitledBorder("Component Types");
        compTypeListPanel.setBorder(compTypeListTitle);

        // comp type list table
        compTypeList = new JTable(new CompTypeResultsTableModel());
        compTypeList.setShowHorizontalLines(true);
        compTypeList.setColumnSelectionAllowed(false);
        compTypeList.setRowSelectionAllowed(true);
        compTypeList.getColumnModel().getColumn(0).setCellRenderer(new CompTypeTableCellRenderer());
    	JScrollPane compTypeListScroller = new JScrollPane(compTypeList, vsbp, hsbp);
        compTypeListPanel.add(compTypeListScroller);
        compTypeList.getColumnModel().getColumn(0).setPreferredWidth(147);
        compTypeList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        // handle row selection in comp type list table
        compTypeList.getSelectionModel().addListSelectionListener(new ListSelectionListener() {
                public void valueChanged(ListSelectionEvent e) {
                    if (e.getValueIsAdjusting()) return;
                    ListSelectionModel lsm = (ListSelectionModel)e.getSource();
                    if (!lsm.isSelectionEmpty()) {
                        int selectedCompTypeResultsRow = lsm.getMinSelectionIndex();
                        List compTypeList;
                        compTypeList = _model.getComponentTypeList();
                        ComponentType selectedComponentType = 
                            (ComponentType)compTypeList.get(selectedCompTypeResultsRow);
                        _model.setSelectedComponentType(selectedComponentType);
                        //document.actionAttributeSearch();
                    }
                }
            });

        // component type action button bar
        JPanel compTypeButtons = new JPanel(new FlowLayout(FlowLayout.LEFT));
        compTypeEditButton = new JButton("Edit...");
        LoginUtil.registerProtectedComponent(compTypeEditButton, adminPrincipal);
        compTypeEditButton.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    JFrame appFrame = 
                        ((DesktopApplication)Application.getApp()).getDesktopFrame();
                    /*
                    IOC ioc = 
                        IOCEditDialog.showDialog(appFrame,null,_model.getSelectedIoc(),false);
                    document.actionSaveIoc(ioc,true);
                    */
                }
            });
        
        compTypeAddButton = new JButton("Add...");
        LoginUtil.registerProtectedComponent(compTypeAddButton, adminPrincipal);
        compTypeAddButton.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    JFrame appFrame = 
                        ((DesktopApplication)Application.getApp()).getDesktopFrame();
                    /*
                    IOC ioc = 
                        IOCEditDialog.showDialog(appFrame,null,null,true);
                    document.actionSaveIoc(ioc,true);
                    */
                }
            });

        compTypeButtons.add(compTypeEditButton);
        compTypeButtons.add(compTypeAddButton);

        // add comp type list and button panels to compTypesPanel
        compTypesPanel.add(compTypeListPanel);
        compTypesPanel.add(compTypeButtons);

        // make the panel used to display data after comp type is selected
        selectedCompTypePanel = makeSelectedCompTypePanelContents();

        // split between iocsPanel and ioc details info on right
        lrSplitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT,
                                     compTypesPanel,
                                     selectedCompTypePanel);
        lrSplitPane.setDividerLocation(150);

        // add split panes to topPanel
        topPanel.add(lrSplitPane,c);

    	getContentPane().add(topPanel);
    }

    /**
     *
     */
    private JPanel makeSelectedCompTypePanelContents() {
        
        // scrollbar policy
        int hsbp = JScrollPane.HORIZONTAL_SCROLLBAR_NEVER;
        int vsbp = JScrollPane.VERTICAL_SCROLLBAR_ALWAYS;

        // comp type details panel containing details of a particular type
        compTypeDetailsPanel = new JPanel();
        compTypeDetailsPanel.setLayout(new BoxLayout(compTypeDetailsPanel,BoxLayout.PAGE_AXIS));

        // ioc attributes panel
        /*
        iocAttrPanel = new JPanel(new BorderLayout());
        TitledBorder iocAttrTitle = BorderFactory.createTitledBorder("IOC Attributes");
        iocAttrPanel.setBorder(iocAttrTitle);
        */

        // ioc attributes table
        /*
        attrList = new JTable(new IOCAttrTableModel());
        attrList.setShowHorizontalLines(true);
        attrList.setRowSelectionAllowed(true);
        attrList.getColumnModel().getColumn(0).setCellRenderer(new IOCAttrNameTableCellRenderer());
        attrList.getColumnModel().getColumn(0).setCellEditor(new IOCAttrNameTableCellEditor());
        attrList.getColumnModel().getColumn(1).setCellRenderer(new IOCAttrValueTableCellRenderer());
        attrList.getColumnModel().getColumn(1).setCellEditor(new IOCAttrValueTableCellEditor());

    	JScrollPane attrListScroller = new JScrollPane(attrList, vsbp, hsbp);
        iocAttrPanel.add(attrListScroller);
        attrList.getColumnModel().getColumn(1).setPreferredWidth(50);
        attrList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        // handle row selection in attribute list table
        attrList.getSelectionModel().addListSelectionListener(new ListSelectionListener() {
                public void valueChanged(ListSelectionEvent e) {
                    if (e.getValueIsAdjusting()) return;
                    ListSelectionModel lsm = (ListSelectionModel)e.getSource();
                    if (!lsm.isSelectionEmpty()) {
                        int selectedAttributeRow = lsm.getMinSelectionIndex();
                        int lastRow = displayAttrList.size() - 1;

                        // remove any incomplete "add attribute" first
                        IOCAttribute possibleIncompleteAttr = _model.getSelectedIocAttribute();
                        if (possibleIncompleteAttr != null &&
                            possibleIncompleteAttr.getId() == null &&
                            selectedAttributeRow != lastRow) {
                            attrList.getCellEditor(lastRow,0).cancelCellEditing();
                            displayAttrList.remove(possibleIncompleteAttr);
                            IOCAttrTableModel iocAttrTableModel = 
                                (IOCAttrTableModel)attrList.getModel();
                            iocAttrTableModel.fireTableRowsDeleted(lastRow,lastRow);
                        }
                            

                        // now initialize new attribute selection
                        IOCAttribute selectedIocAttr = 
                            (IOCAttribute)displayAttrList.get(selectedAttributeRow);
                        _model.setSelectedIocAttribute(selectedIocAttr);
                    }
                }
            });

        // attributes action button bar
        JPanel attrButtons = new JPanel(new FlowLayout(FlowLayout.LEFT));
        
        attrAddButton = new JButton("Add");
        LoginUtil.registerProtectedComponent(attrAddButton, adminPrincipal);
        attrAddButton.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    // create new attribute, but just display it for now
                    IOCAttribute newAttr = new IOCAttribute();
                    newAttr.setIoc(_model.getSelectedIoc());
                    _model.setSelectedIocAttribute(newAttr);
                    displayAttrList.add(newAttr);
                    int row = displayAttrList.size() - 1;
                    IOCAttrTableModel iocAttrTableModel = 
                        (IOCAttrTableModel)attrList.getModel();
                    iocAttrTableModel.fireTableRowsInserted(row, row);
                    attrList.editCellAt(row,0);
                }
            });

        attrRemoveButton = new JButton("Remove");
        LoginUtil.registerProtectedComponent(attrRemoveButton, adminPrincipal);
        attrRemoveButton.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    if (_model.getSelectedIocAttribute() != null) {
                        displayAttrList.remove(_model.getSelectedIocAttribute());
                        document.actionDeleteIocAttribute(_model.getSelectedIocAttribute());
                    }
                }
            });

        attrButtons.add(attrAddButton);
        attrButtons.add(attrRemoveButton);
        */
        

        // add attribute subpanel to comp type details panel
        //compTypeDetailsPanel.add(iocAttrPanel);
        //iocDetailsPanel.add(attrButtons);

        return compTypeDetailsPanel;
    }


    /**
     * Update graphical view of ComponentTypeModel data based on event type.
     *
     * @param event type of data model change
     */
    public void updateView(ComponentTypeModelEvent event) {

        switch(event.getType()) {

        case ComponentTypeModelEvent.NEW_COMPONENT_TYPE_LIST: {
            CompTypeResultsTableModel tableModel = 
                (CompTypeResultsTableModel)compTypeList.getModel();

            // request that comp type results table update itself
            tableModel.fireTableDataChanged();
            if (_model.getComponentTypeList() != null && 
                _model.getComponentTypeList().size() > 0 &&
                compTypeList.getSelectedRow() == -1) {
                compTypeList.setRowSelectionInterval(0,0);  // preselect row 0
            }            
            break;
        }
            
        default: {}
        }
            
    }

    /********************************************************************
     * Inner classes supporting Swing components
     ********************************************************************/

    /**
     * Provides methods to get comp type results table cell data from
     * actual <code>ComponentTypeModel</code>. 
     */
    class CompTypeResultsTableModel extends AbstractTableModel {
        private String[] columnNames = {"Component Type"};
        public int getColumnCount() {
            return columnNames.length;
        }
        public int getRowCount() {
            int size = 0;
            if (_model.getComponentTypeList() == null)
                return size;
            size = _model.getComponentTypeList().size();
            return size;
        }
        public String getColumnName(int col) {
            return columnNames[col];
        }
        public Object getValueAt(int row, int col) {
            List compTypeList = null;

            compTypeList = _model.getComponentTypeList();

            if (compTypeList != null) {            
                ComponentType compType = (ComponentType)compTypeList.get(row);
                return compType.getComponentTypeName();
            } else {
                return null;
            }
        }
        public Class getColumnClass(int c) {
            return getValueAt(0,c).getClass();
        }
    }

    /**
     * Customized table cell renderer used for comp type table. 
     */
    class CompTypeTableCellRenderer extends DefaultTableCellRenderer {
        public Component getTableCellRendererComponent (JTable table,
                                                        Object value,
                                                        boolean selected,
                                                        boolean focused,
                                                        int row, int column) {
            setEnabled(table==null || table.isEnabled());
            if (column == 0) {
                if (!selected)
                    setBackground(Color.white);
                else
                    setBackground(new Color(0xCC,0xCC,0xFF));

                setToolTipText(null);
                setForeground(Color.blue);
            }
            super.getTableCellRendererComponent(table,value,selected,focused,row,column);
            return this;
        }
    }
}
