/*
   Copyright (c) 2004-2005 The University of Chicago, as Operator
   of Argonne National Laboratory.
*/
package gov.anl.aps.irmis.apps.ioc.cfw;

import java.io.File;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Iterator;
import java.util.Enumeration;
import java.util.Set;
import java.util.EventObject;
import java.util.logging.*;

import java.awt.*;
import java.awt.event.*;

import javax.swing.*;
import javax.swing.table.*;
import javax.swing.border.TitledBorder;
import javax.swing.event.ListSelectionListener;
import javax.swing.event.ListSelectionEvent;

import gov.sns.application.*;

// irmisBase db imports
import gov.anl.aps.irmis.persistence.pv.IOC;
import gov.anl.aps.irmis.persistence.pv.IOCAttribute;
import gov.anl.aps.irmis.persistence.pv.IOCAttributeType;

// for registering protected gui components
import gov.anl.aps.irmis.login.LoginUtil;
import gov.anl.aps.irmis.login.RolePrincipal;

/**
 * Primary GUI for IRMIS IOC application. All window layout, and Swing event
 * listeners are done here. Business logic work is delegated to <code>IOCDocument</code>,
 * which in turn requests that the IOCModel notify us here of changes to the data. In short,
 * we listen for <code>IOCModelEvent</code> here.
 */
public class IOCWindow extends XalInternalWindow {

	/** The main model for the document */
	final protected IOCModel _model;

    final protected IOCDocument document;

    static RolePrincipal[] adminPrincipal = { new RolePrincipal("admin") };

    // ordered list of attribute results for selected IOC
    private List displayAttrList;

	// Swing GUI components
    private JPanel topPanel;
    private JPanel iocsPanel;
    private JPanel iocListPanel;
    private JPanel selectedIocPanel;
    private JPanel iocDetailsPanel;
    private JPanel iocAttrPanel;

    private JSplitPane lrSplitPane;

    private JTable iocList;
    private JTable attrList;

    private JButton iocEditButton;
    private JButton iocAddButton;
    private JButton attrEditButton;
    private JButton attrAddButton;
    private JButton attrRemoveButton;
	
	/** 
	 * Creates a new instance of ComponentWindow
	 * @param aDocument The document for this window
	 */
    public IOCWindow(final XalInternalDocument aDocument) {
        super(aDocument);
        document = (IOCDocument)aDocument;

		_model = document.getModel();

        // make IOCWindow a listener for changes in IOCModel
		_model.addIOCModelListener( new IOCModelListener() {
                public void modified(IOCModelEvent e) {
                    updateView(e);
                }
            });

        // initial application window size
        setSize(420, 205);

        // build contents
		makeContents();
    }

    /**
     * Top-level method to build up Swing GUI components.
     */
    public void makeContents() {

        // scrollbar policy
        int hsbp = JScrollPane.HORIZONTAL_SCROLLBAR_NEVER;
        int vsbp = JScrollPane.VERTICAL_SCROLLBAR_ALWAYS;

        // topmost panel
        topPanel = new JPanel();
        topPanel.setLayout(new GridBagLayout());
        GridBagConstraints c = new GridBagConstraints();
        c.weightx = 1.0;
        c.weighty = 1.0;
        c.fill = GridBagConstraints.BOTH;
        c.gridx = 0;
        c.gridy = 0;


        iocsPanel = new JPanel();
        iocsPanel.setLayout(new BoxLayout(iocsPanel,BoxLayout.PAGE_AXIS));

        // ioc list panel
        iocListPanel = new JPanel(new BorderLayout());
        TitledBorder iocListTitle = BorderFactory.createTitledBorder("IOCs");
        iocListPanel.setBorder(iocListTitle);

        // ioc list table
        iocList = new JTable(new IOCResultsTableModel());
        iocList.setShowHorizontalLines(true);
        iocList.setColumnSelectionAllowed(false);
        iocList.setRowSelectionAllowed(true);
        iocList.getColumnModel().getColumn(0).setCellRenderer(new IOCTableCellRenderer());
    	JScrollPane iocListScroller = new JScrollPane(iocList, vsbp, hsbp);
        iocListPanel.add(iocListScroller);
        iocList.getColumnModel().getColumn(0).setPreferredWidth(147);
        iocList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        // handle row selection in ioc list table
        iocList.getSelectionModel().addListSelectionListener(new ListSelectionListener() {
                public void valueChanged(ListSelectionEvent e) {
                    if (e.getValueIsAdjusting()) return;
                    ListSelectionModel lsm = (ListSelectionModel)e.getSource();
                    if (!lsm.isSelectionEmpty()) {
                        int selectedIOCResultsRow = lsm.getMinSelectionIndex();
                        List iocList;
                        iocList = _model.getIocList();
                        IOC selectedIoc = (IOC)iocList.get(selectedIOCResultsRow);
                        _model.setSelectedIoc(selectedIoc);
                        document.actionAttributeSearch();
                    }
                }
            });

        // ioc action button bar
        JPanel iocButtons = new JPanel(new FlowLayout(FlowLayout.LEFT));
        iocEditButton = new JButton("Edit...");
        LoginUtil.registerProtectedComponent(iocEditButton, adminPrincipal);
        iocEditButton.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    JFrame appFrame = 
                        ((DesktopApplication)Application.getApp()).getDesktopFrame();
                    IOC ioc = 
                        IOCEditDialog.showDialog(appFrame,null,_model.getSelectedIoc(),false);
                    document.actionSaveIoc(ioc,true);
                }
            });
        
        iocAddButton = new JButton("Add...");
        LoginUtil.registerProtectedComponent(iocAddButton, adminPrincipal);
        iocAddButton.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    JFrame appFrame = 
                        ((DesktopApplication)Application.getApp()).getDesktopFrame();
                    IOC ioc = 
                        IOCEditDialog.showDialog(appFrame,null,null,true);
                    document.actionSaveIoc(ioc,true);
                }
            });

        iocButtons.add(iocEditButton);
        iocButtons.add(iocAddButton);

        // add ioc list and button panels to iocsPanel
        iocsPanel.add(iocListPanel);
        iocsPanel.add(iocButtons);

        // make the panel used to display data after ioc is selected
        selectedIocPanel = makeSelectedIocPanelContents();

        // split between iocsPanel and ioc details info on right
        lrSplitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT,iocsPanel,selectedIocPanel);
        lrSplitPane.setDividerLocation(150);

        // add split panes to topPanel
        topPanel.add(lrSplitPane,c);

    	getContentPane().add(topPanel);
    }

    private JPanel makeSelectedIocPanelContents() {
        
        // scrollbar policy
        int hsbp = JScrollPane.HORIZONTAL_SCROLLBAR_NEVER;
        int vsbp = JScrollPane.VERTICAL_SCROLLBAR_ALWAYS;

        // ioc details panel containing ioc attribute info
        iocDetailsPanel = new JPanel();
        iocDetailsPanel.setLayout(new BoxLayout(iocDetailsPanel,BoxLayout.PAGE_AXIS));

        // ioc attributes panel
        iocAttrPanel = new JPanel(new BorderLayout());
        TitledBorder iocAttrTitle = BorderFactory.createTitledBorder("IOC Attributes");
        iocAttrPanel.setBorder(iocAttrTitle);

        // ioc attributes table
        attrList = new JTable(new IOCAttrTableModel());
        attrList.setShowHorizontalLines(true);
        attrList.setRowSelectionAllowed(true);
        attrList.getColumnModel().getColumn(0).setCellRenderer(new IOCAttrNameTableCellRenderer());
        attrList.getColumnModel().getColumn(0).setCellEditor(new IOCAttrNameTableCellEditor());
        attrList.getColumnModel().getColumn(1).setCellRenderer(new IOCAttrValueTableCellRenderer());
        attrList.getColumnModel().getColumn(1).setCellEditor(new IOCAttrValueTableCellEditor());

    	JScrollPane attrListScroller = new JScrollPane(attrList, vsbp, hsbp);
        iocAttrPanel.add(attrListScroller);
        attrList.getColumnModel().getColumn(1).setPreferredWidth(50);
        attrList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        // handle row selection in attribute list table
        attrList.getSelectionModel().addListSelectionListener(new ListSelectionListener() {
                public void valueChanged(ListSelectionEvent e) {
                    if (e.getValueIsAdjusting()) return;
                    ListSelectionModel lsm = (ListSelectionModel)e.getSource();
                    if (!lsm.isSelectionEmpty()) {
                        int selectedAttributeRow = lsm.getMinSelectionIndex();
                        int lastRow = displayAttrList.size() - 1;

                        // remove any incomplete "add attribute" first
                        IOCAttribute possibleIncompleteAttr = _model.getSelectedIocAttribute();
                        if (possibleIncompleteAttr != null &&
                            possibleIncompleteAttr.getId() == null &&
                            selectedAttributeRow != lastRow) {
                            attrList.getCellEditor(lastRow,0).cancelCellEditing();
                            displayAttrList.remove(possibleIncompleteAttr);
                            IOCAttrTableModel iocAttrTableModel = 
                                (IOCAttrTableModel)attrList.getModel();
                            iocAttrTableModel.fireTableRowsDeleted(lastRow,lastRow);
                        }
                            

                        // now initialize new attribute selection
                        IOCAttribute selectedIocAttr = 
                            (IOCAttribute)displayAttrList.get(selectedAttributeRow);
                        _model.setSelectedIocAttribute(selectedIocAttr);
                    }
                }
            });

        // attributes action button bar
        JPanel attrButtons = new JPanel(new FlowLayout(FlowLayout.LEFT));
        
        attrAddButton = new JButton("Add");
        LoginUtil.registerProtectedComponent(attrAddButton, adminPrincipal);
        attrAddButton.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    // create new attribute, but just display it for now
                    IOCAttribute newAttr = new IOCAttribute();
                    newAttr.setIoc(_model.getSelectedIoc());
                    _model.setSelectedIocAttribute(newAttr);
                    displayAttrList.add(newAttr);
                    int row = displayAttrList.size() - 1;
                    IOCAttrTableModel iocAttrTableModel = 
                        (IOCAttrTableModel)attrList.getModel();
                    iocAttrTableModel.fireTableRowsInserted(row, row);
                    attrList.editCellAt(row,0);
                }
            });

        attrRemoveButton = new JButton("Remove");
        LoginUtil.registerProtectedComponent(attrRemoveButton, adminPrincipal);
        attrRemoveButton.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    if (_model.getSelectedIocAttribute() != null) {
                        displayAttrList.remove(_model.getSelectedIocAttribute());
                        document.actionDeleteIocAttribute(_model.getSelectedIocAttribute());
                    }
                }
            });

        attrButtons.add(attrAddButton);
        attrButtons.add(attrRemoveButton);
        

        // add attribute subpanel to ioc details panel
        iocDetailsPanel.add(iocAttrPanel);
        iocDetailsPanel.add(attrButtons);

        return iocDetailsPanel;
    }

    /**
     * Update graphical view of IOCModel data based on event type.
     *
     * @param event type of data model change
     */
    public void updateView(IOCModelEvent event) {

        switch(event.getType()) {
            
        case IOCModelEvent.NEW_IOC_LIST: {
            IOCResultsTableModel tableModel = (IOCResultsTableModel)iocList.getModel();
            // request that ioc results table update itself
            tableModel.fireTableDataChanged();
            if (_model.getIocList() != null && _model.getIocList().size() > 0 &&
                iocList.getSelectedRow() == -1) {
                iocList.setRowSelectionInterval(0,0);  // preselect row 0
            }
            break;
        }

        case IOCModelEvent.NEW_IOC_ATTR_LIST: {
            IOC selectedIoc = _model.getSelectedIoc();
            // convert attribute set to an ordered list
            Set attributes = selectedIoc.getAttributes();
            Iterator attrIt = attributes.iterator();
            displayAttrList = new ArrayList();
            while (attrIt.hasNext()) {
                IOCAttribute attr = (IOCAttribute)attrIt.next();
                displayAttrList.add(attr);
            }
            Collections.sort(displayAttrList);
            IOCAttrTableModel iocAttrTableModel = 
                (IOCAttrTableModel)attrList.getModel();
            iocAttrTableModel.fireTableDataChanged();
            break;
        }

        case IOCModelEvent.NEW_IOC_SELECTION: {
            List iocs = _model.getIocList();
            IOC selectedIoc = _model.getSelectedIoc();
            int row = -1;
            if (iocs != null) {
                Iterator iocIt = iocs.iterator();
                while (iocIt.hasNext()) {
                    row++;
                    IOC ioc = (IOC)iocIt.next();
                    if (ioc.getIocName().equals(selectedIoc.getIocName())) {
                        break;
                    }
                }                
            }

            // weird sequence to get selection changed and visible
            iocList.changeSelection(row,0,false,false);
            Rectangle cell = iocList.getCellRect(row,0,true);
            iocList.scrollRectToVisible(cell);
            break;
        }

        case IOCModelEvent.IOC_SAVED: {
            IOCResultsTableModel tableModel = (IOCResultsTableModel)iocList.getModel();
            // request that ioc results table update itself
            tableModel.fireTableDataChanged();            
            break;
        }

        case IOCModelEvent.IOC_ATTRIBUTE_DELETED: {
            IOCAttrTableModel iocAttrTableModel = 
                (IOCAttrTableModel)attrList.getModel();
            iocAttrTableModel.fireTableDataChanged();
            break;
        }

        default: {}
        }
            
    }

    /********************************************************************
     * Inner classes supporting Swing components
     ********************************************************************/

    /**
     * Provides methods to get ioc results table cell data from
     * actual <code>IOCModel</code>. 
     */
    class IOCResultsTableModel extends AbstractTableModel {
        private String[] columnNames = {"IOC Name"};
        public int getColumnCount() {
            return columnNames.length;
        }
        public int getRowCount() {
            int size = 0;
            if (_model.getIocList() == null)
                return size;
            size = _model.getIocList().size();
            return size;
        }
        public String getColumnName(int col) {
            return columnNames[col];
        }
        public Object getValueAt(int row, int col) {
            List iocList = null;

            iocList = _model.getIocList();

            if (iocList != null) {            
                IOC ioc = (IOC)iocList.get(row);
                return ioc.getIocName();
            } else {
                return null;
            }
        }
        public Class getColumnClass(int c) {
            return getValueAt(0,c).getClass();
        }
    }

    /**
     * Provides methods to get ioc attribute table cell data from actual
     * <code>IOCModel</code>. See Swing tutorial.
     */
    class IOCAttrTableModel extends AbstractTableModel {
        private String[] columnNames = {"Name","Value"};
        public int getColumnCount() {
            return columnNames.length;
        }
        public int getRowCount() {
            int size = 0;
            if (displayAttrList != null) {
                size = displayAttrList.size();
            }
            return size;
        }
        public String getColumnName(int col) {
            return columnNames[col];
        }
        public Object getValueAt(int row, int col) {

            if (displayAttrList != null) {
                IOCAttribute attr = (IOCAttribute)displayAttrList.get(row);
                if (col == 0) {
                    if (attr.getAttributeType() == null)
                        return "";
                    else
                        return attr.getAttributeType().getAttributeType();
                } else {
                    if (attr.getAttributeValue() == null)
                        return "";
                    else
                        return attr.getAttributeValue();
                }
            } else {
                return null;
            }

        }

        // called after edit of attribute cell is complete
        public void setValueAt(Object value, int row, int col) {
            String strValue = (String)value;

            if (displayAttrList != null) {
                IOCAttribute attr = (IOCAttribute)displayAttrList.get(row);
                if (col == 0) {  // attribute name

                    // look for existing attribute type object
                    List attrTypeList = _model.getIocAttributeTypeList();
                    Iterator attrTypeIt = attrTypeList.iterator();
                    IOCAttributeType attrType = null;
                    boolean isNewAttrType = true;
                    while (attrTypeIt.hasNext()) {
                        IOCAttributeType existingAttrType = (IOCAttributeType)attrTypeIt.next();
                        if (existingAttrType.getAttributeType().equals(strValue)) {
                            attrType = existingAttrType;
                            isNewAttrType = false;
                            break;
                        }
                    }
                    // make new one if we didn't find existing
                    if (attrType == null) {
                        attrType = new IOCAttributeType();
                        attrType.setAttributeType(strValue);
                    }

                    // add attribute type to new attribute
                    attr.setAttributeType(attrType);
                    // final step of actually adding attribute to ioc object
                    attr.getIoc().addAttribute(attr);
                    document.actionSaveIoc(attr.getIoc(),false);

                    // add attrType to model's getIocAttributeTypeList if it's new
                    if (isNewAttrType) {
                        attrTypeList.add(attrType);
                        Collections.sort(attrTypeList);
                    }

                } else if (col == 1) {  // attribute value
                    attr.setAttributeValue(strValue);
                    // save attribute by requesting save of parent ioc
                    document.actionSaveIoc(attr.getIoc(),false);
                }
            }
        }

        public Class getColumnClass(int c) {
            return getValueAt(0,c).getClass();
        }

        public boolean isCellEditable(int row, int col) {
            if (col == 0) {
                // only permit editing name once for newly added rows
                String cellValue = (String)getValueAt(row,col);
                if (cellValue.length() == 0 &&
                    LoginUtil.isPermitted(adminPrincipal))
                    return true;
                else
                    return false;
                
            } else if (col == 1) {
                // permit editing value only if user logged in with sufficient permissions
                if (LoginUtil.isPermitted(adminPrincipal))
                    return true;
                else 
                    return false;
            }
            return false;
        }
    }
    
    /**
     * Customized table cell renderer used for ioc table. It colors
     * the cell background and text foreground based on whether ioc is active
     * or not.
     */
    class IOCTableCellRenderer extends DefaultTableCellRenderer {
        public Component getTableCellRendererComponent (JTable table,
                                                        Object value,
                                                        boolean selected,
                                                        boolean focused,
                                                        int row, int column) {
            setEnabled(table==null || table.isEnabled());
            if (column == 0) {
                List iocList = _model.getIocList();
                if (iocList != null) {
                    IOC ioc = (IOC)iocList.get(row);
                    if (!ioc.getActive()) {
                        setBackground(Color.pink);
                        setToolTipText("inactive");
                    } else {
                        if (!selected)
                            setBackground(Color.white);
                        else
                            setBackground(new Color(0xCC,0xCC,0xFF));
                        setToolTipText(null);
                    }
                    setForeground(Color.blue);
                }
            }
            super.getTableCellRendererComponent(table,value,selected,focused,row,column);
            return this;
        }
    }

    /**
     * Cell renderer for name column of ioc attribute table.
     */
    class IOCAttrNameTableCellRenderer 
        implements TableCellRenderer {
        
        public Component getTableCellRendererComponent(JTable table, 
                                                       Object value, 
                                                       boolean isSelected,
                                                       boolean hasFocus,
                                                       int row, int col) {
            JTextArea textArea = new JTextArea();

            if (value != null)
                textArea.setText(value.toString());

            if (isSelected)
                textArea.setBackground(new Color(0xCC,0xCC,0xFF));
            else
                textArea.setBackground(Color.white);
            
            return textArea;
        }
    }

    /**
     * Cell editor for name column of ioc attribute table.
     */
    class IOCAttrNameTableCellEditor extends AbstractCellEditor 
        implements TableCellEditor {

        JComboBox cb = new JComboBox();

        // called when cell value edited by user
        public Component getTableCellEditorComponent(JTable table,
                                                     Object value, 
                                                     boolean isSelected,
                                                     int row, int col) {

            String valueStr = (String)value;
            // set up combo box here
            List attrTypesList = _model.getIocAttributeTypeList();
            if (attrTypesList != null) {
                cb.removeAllItems();
                cb.setEditable(true);
                Iterator attrTypeIt = attrTypesList.iterator();
                while (attrTypeIt.hasNext()) {
                    IOCAttributeType attrType = (IOCAttributeType)attrTypeIt.next();
                    String typeString = attrType.getAttributeType();
                    cb.addItem(typeString);
                }
                
                // autosize the row height to fit combo box better
                int desiredHeight = (int) cb.getPreferredSize().getHeight();
                if (desiredHeight > table.getRowHeight(row))
                    table.setRowHeight(row,desiredHeight);
            }

            return cb;
        }

        // called when editing is complete. must return new value
        // to be stored in the cell.
        public Object getCellEditorValue() {
            String selectedValue = (String)cb.getSelectedItem();
            return selectedValue;
        }
        
        public boolean isCellEditable(EventObject e) {
            return true;
        }

        public boolean stopCellEditing() {
            // prevent user from choosing already defined attribute
            String newAttrType = (String)getCellEditorValue();
            Iterator attrIt = displayAttrList.iterator();
            while (attrIt.hasNext()) {
                IOCAttribute existingAttr = (IOCAttribute)attrIt.next();
                if ((existingAttr.getAttributeType() != null &&
                    existingAttr.getAttributeType().getAttributeType().equals(newAttrType)) ||
                    newAttrType.equals("** duplicate **")) {
                    cb.setSelectedItem(new String("** duplicate **"));
                    return false;
                }
            }

            return super.stopCellEditing();
        }

    }

    /**
     * Cell renderer for value column of ioc attribute table.
     */
    class IOCAttrValueTableCellRenderer 
        implements TableCellRenderer {
        
        public Component getTableCellRendererComponent(JTable table, 
                                                       Object value, 
                                                       boolean isSelected,
                                                       boolean hasFocus,
                                                       int row, int col) {
            JTextArea textArea = new JTextArea();
            // modify properties of text area here
            textArea.setLineWrap(true);
            textArea.setWrapStyleWord(true);
            textArea.setToolTipText("Click to edit (you must be logged in)");

            if (value != null)
                textArea.setText(value.toString());

            // autosize the row height to fit text exactly
            textArea.setSize(table.getColumnModel().getColumn(col).getWidth(), Integer.MAX_VALUE);
            int desiredHeight = (int) textArea.getPreferredSize().getHeight();
            if (desiredHeight > table.getRowHeight(row))
                table.setRowHeight(row,desiredHeight);

            if (isSelected)
                textArea.setBackground(new Color(0xCC,0xCC,0xFF));
            else
                textArea.setBackground(Color.white);
            
            return textArea;
        }
    }

    /**
     * Cell editor for value column of ioc attribute table.
     */
    class IOCAttrValueTableCellEditor extends AbstractCellEditor 
        implements TableCellEditor {

        JTextArea textArea = new JTextArea();
        // called when cell value edited by user
        public Component getTableCellEditorComponent(JTable table,
                                                     Object value, 
                                                     boolean isSelected,
                                                     int row, int col) {

            // modify properties of text area here
            textArea.setLineWrap(true);
            textArea.setWrapStyleWord(true);
            textArea.setEditable(true);
            
            // put the data in the text field
            if (value != null)
                textArea.setText(value.toString());
            
            // autosize the row height to fit text exactly
            textArea.setSize(table.getColumnModel().getColumn(col).getWidth(), Integer.MAX_VALUE);
            int desiredHeight = (int) textArea.getPreferredSize().getHeight();
            if (desiredHeight > table.getRowHeight(row))
                table.setRowHeight(row,desiredHeight);
            
            //textArea.selectAll();
            //textArea.setBackground(Color.white);

            return textArea;
        }

        // called when editing is complete. must return new value
        // to be stored in the cell.
        public Object getCellEditorValue() {
            return textArea.getText();
        }

        public boolean isCellEditable(EventObject e) {
            /* here is a way to implement double-click editing requests
            if (e instanceof MouseEvent) {
                MouseEvent me = (MouseEvent)e;
                if (me.getClickCount() == 2)
                    return true;
                else
                    return false;
            } else {
                return false;
            }
            */
            return true;
        }

    }

}
